// monoral double valued wav-file class
// Copyright (C) 1994,2001 Yukinori NAGATANI
// nagatani@yukawa.kyoto-u.ac.jp

#include <sys/types.h>	// write
#include <sys/uio.h>	// write
#include <unistd.h>	// write

typedef	unsigned long int	u_32;
typedef	unsigned short int	u_16;

//////////////////////////////////////////////////////////////////////
// Define values for Microsoft WAVE format
const u_32	RIFF =	0x46464952;
const u_32	WAVE =	0x45564157;
const u_32	FMT  =	0x20746D66;
const u_32	DATA =	0x61746164;
const u_32	PCM_CODE	= 1;
const u_32	WAVE_MONO	= 1;
const u_32	WAVE_STEREO	= 2;

class WaveFileHeader {
public:
	u_32	main_chunk;  // 'RIFF'
	u_32	length;      // filelen
	u_32	chunk_type;  // 'WAVE'
	
	u_32	sub_chunk;   // 'fmt '
	u_32	sc_len;      // length of sub_chunk, =16
	u_16	format;      // should be 1 for PCM-code
	u_16	modus;       // 1 Mono, 2 Stereo
	u_32	sample_fq;   // frequence of sample
	u_32	byte_p_sec;  // byte per second
	u_16	byte_p_spl;  // samplesize; 1 or 2 bytes
	u_16	bit_p_spl;   // 8, 12 or 16 bit
	
	u_32	data_chunk;  // 'data'
	u_32	data_length; // samplecount
	
	WaveFileHeader (u_32 nDataLength = 0) {
		setAll (nDataLength);
	}
	void setAll (u_32 nDataLength) {
		main_chunk	= RIFF;
		length		= nDataLength + sizeof (WaveFileHeader);
		chunk_type	= WAVE;
		
		sub_chunk	= FMT;
		sc_len		= 16;		// length of sub_chunk
		format		= PCM_CODE;	// should be 1 for PCM-code
		modus		= WAVE_MONO;	// 1 Mono, 2 Stereo
		sample_fq	= 44100;	// frequence of sample (Hz)
		byte_p_sec	= 44100 * 2;
		byte_p_spl	= 2;		// 1 or 2 bytes
		bit_p_spl	= 16;		// bit
		
		data_chunk	= DATA;
		data_length	= nDataLength; // samplecount
	}
	void setStereo (bool bIsStereo) {
		if (bIsStereo)
			modus = WAVE_STEREO;	// 1 Mono, 2 Stereo
		else
			modus = WAVE_MONO;
	}
	bool bIsStereo (void) const {
		return	(modus == WAVE_STEREO);
	}
	void setDataLength (u_32 nDataLength) {
		length		= nDataLength + sizeof (WaveFileHeader);
		data_length	= nDataLength; // samplecount
	}
	void setFreq (int nFreq) {
		sample_fq	= nFreq;	// frequence of sample (Hz)
		byte_p_sec	= nFreq / 8;
	}
	int write (int fd = 1) const {
		return	::write (fd, this, sizeof (WaveFileHeader));
	}
	int read (int fd = 0) {
		return ::read (fd, this, sizeof (WaveFileHeader));
	}
	static void show_chars (u_32 chunk_name) {
		char	* buf = (char *) & chunk_name;
		fprintf (stderr, "%c%c%c%c\n",
			 buf [0], buf [1], buf [2], buf [3]);
	}
	void show () const {
		
		fprintf (stderr, "main_chunk:");
		show_chars (main_chunk);
		fprintf (stderr, "file length:%ld\n", length);
		fprintf (stderr, "chunk_type:");
		show_chars (chunk_type);
		
		fprintf (stderr, "length of sub_chunk:%ld\n", sc_len);
		fprintf (stderr, "format [PCM should=1]:%d\n", format);	
		fprintf (stderr, "modus [1 Mono, 2 Stereo ]:%d\n", modus);
		fprintf (stderr, "sample_fq:%ld\n", sample_fq);
		fprintf (stderr, "byte_p_sec_fq:%ld\n", byte_p_sec);
		fprintf (stderr, "byte_p_spl:%d\n", byte_p_spl);
		fprintf (stderr, "bit_p_spl [8, 12, 16 bit]:%d\n", bit_p_spl);
		
		fprintf (stderr, "data_chunk:");
		show_chars (data_chunk);
		fprintf (stderr, "data_length:%ld\n", data_length);
		
		fprintf (stderr, "size of header:%ld\n",
			 sizeof (WaveFileHeader));
	}
};

//////////////////////////////////////////////////////////////////////

class PcmUnit {
public:
	unsigned char	Low;
	unsigned char	High;
	
	PcmUnit (int nLevel = 0) {
		set (nLevel);
	}
	void set (int nLevel) {	// -32768 < nLevel < 32767
		if (nLevel < -32768)
			nLevel = -32768;
		if (nLevel > +32767)
			nLevel = +32767;
		nLevel  += 256*128;
		Low = (unsigned char) (nLevel & 0x00ff);
		High = (unsigned char) (nLevel / 256 - 128);
	}
	void set (double rLevel) { // -1.0 < rLevel < 1.0
		set ((int) (32767.0 * rLevel));
	}
	double getReal (void) const {
		return	(double)
			(Low + 256 * (signed int) (signed char) High)
			/ 32767.0;
	}
	int write (int fd = 1) const {
		return	::write (fd, this, sizeof (PcmUnit));
	}
	int read (int fd = 0) {
		return	::read (fd, this, sizeof (PcmUnit));
	}
};

//////////////////////////////////////////////////////////////////////

typedef	double	Real;

class RealWave {
public:
	size_t		size,	sizeRec;
	Real		* prBuffer;
	WaveFileHeader	Header;
	void malloc (void) {
		if (prBuffer != NULL)
			delete	prBuffer;
		if (size == 0) {
			prBuffer = NULL;
		} else {
			prBuffer = new Real [size];
			if (prBuffer == NULL) {
				fprintf (stderr, "Cannot Malloc.\n");
				exit (1);
			}
		}
		for (int n = 0 ; n < size ; n++)
			prBuffer [n] = 0.0;
	}
	RealWave (size_t size = 0) {
		this->size = size;
		this->sizeRec = 0;
		prBuffer = NULL;
		malloc ();
		Header.setAll (0);
	}
	size_t read ()
	{
		Header.read ();
//		Header.show ();
		bool bSt2Mono = (Header.modus == WAVE_MONO ? false : true);
		if (bSt2Mono) {
			fprintf (stderr, "Convert Stereo L -> Monoral.\n");
			Header.setStereo (false);
			Header.byte_p_sec /= 2;
			Header.byte_p_spl /= 2;
		} else {
			fprintf (stderr, "Monoral -> Monoral.\n");
		}
		
		PcmUnit	U;
		size_t	n;
		for (n = 0 ; n < size && U.read () > 0 ; n++) {
			prBuffer [n] = U.getReal ();
			if (bSt2Mono)
				U.read ();
		}
		return	sizeRec = n;
	}
	void write (Real rVolume = 1.0)
	{
		// Put Wav Header
		Header.setDataLength (sizeRec * 2);
		Header.setStereo (false);
		Header.write ();
		
		PcmUnit	U;
		for (size_t n = 0 ; n < sizeRec ; n++) {
			U.set (prBuffer[n] * rVolume);
			U.write ();
		}
	}
};

//////////////////////////////////////////////////////////////////////

#ifdef 0
void PutMuLaw( double f, FILE *fOut);
double dMuLaw2Double( unsigned char c);
void PutAuHeader(void);
#endif

